@@ -1,7 +1,7 @@
 /*
  * irc-completion.c - completion for IRC commands
  *
- * Copyright (C) 2003-2015 Sébastien Helleu <flashcode@flashtux.org>
+ * Copyright (C) 2003-2021 Sébastien Helleu <flashcode@flashtux.org>
  *
  * This file is part of WeeChat, the extensible chat client.
  *
@@ -16,7 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with WeeChat.  If not, see <http://www.gnu.org/licenses/>.
+ * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <stdlib.h>
@@ -31,6 +31,7 @@
 #include "irc-completion.h"
 #include "irc-config.h"
 #include "irc-ignore.h"
+#include "irc-modelist.h"
 #include "irc-nick.h"
 #include "irc-notify.h"
 #include "irc-server.h"
@@ -41,20 +42,22 @@
  */
 
 int
-irc_completion_server_cb (void *data, const char *completion_item,
+irc_completion_server_cb (const void *pointer, void *data,
+                          const char *completion_item,
                           struct t_gui_buffer *buffer,
                           struct t_gui_completion *completion)
 {
     IRC_BUFFER_GET_SERVER(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
     if (ptr_server)
     {
-        weechat_hook_completion_list_add (completion, ptr_server->name,
-                                          0, WEECHAT_LIST_POS_SORT);
+        weechat_completion_list_add (completion, ptr_server->name,
+                                     0, WEECHAT_LIST_POS_SORT);
     }
 
     return WEECHAT_RC_OK;
@@ -65,20 +68,22 @@ irc_completion_server_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_server_nick_cb (void *data, const char *completion_item,
+irc_completion_server_nick_cb (const void *pointer, void *data,
+                               const char *completion_item,
                                struct t_gui_buffer *buffer,
                                struct t_gui_completion *completion)
 {
     IRC_BUFFER_GET_SERVER(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
     if (ptr_server && ptr_server->nick)
     {
-        weechat_hook_completion_list_add (completion, ptr_server->nick,
-                                          1, WEECHAT_LIST_POS_SORT);
+        weechat_completion_list_add (completion, ptr_server->nick,
+                                     1, WEECHAT_LIST_POS_SORT);
     }
 
     return WEECHAT_RC_OK;
@@ -89,30 +94,38 @@ irc_completion_server_nick_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_server_channels_cb (void *data, const char *completion_item,
+irc_completion_server_channels_cb (const void *pointer, void *data,
+                                   const char *completion_item,
                                    struct t_gui_buffer *buffer,
                                    struct t_gui_completion *completion)
 {
-    struct t_irc_channel *ptr_channel;
+    struct t_irc_channel *ptr_channel2;
 
-    IRC_BUFFER_GET_SERVER(buffer);
+    IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
-    (void) buffer;
 
     if (ptr_server)
     {
-        for (ptr_channel = ptr_server->channels; ptr_channel;
-             ptr_channel = ptr_channel->next_channel)
+        for (ptr_channel2 = ptr_server->channels; ptr_channel2;
+             ptr_channel2 = ptr_channel2->next_channel)
         {
-            if (ptr_channel->type == IRC_CHANNEL_TYPE_CHANNEL)
+            if (ptr_channel2->type == IRC_CHANNEL_TYPE_CHANNEL)
             {
-                weechat_hook_completion_list_add (completion, ptr_channel->name,
-                                                  0, WEECHAT_LIST_POS_SORT);
+                weechat_completion_list_add (completion, ptr_channel2->name,
+                                             0, WEECHAT_LIST_POS_SORT);
             }
         }
+
+        /* add current channel first in list */
+        if (ptr_channel)
+        {
+            weechat_completion_list_add (completion, ptr_channel->name,
+                                         0, WEECHAT_LIST_POS_BEGINNING);
+        }
     }
 
     return WEECHAT_RC_OK;
@@ -123,7 +136,8 @@ irc_completion_server_channels_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_server_privates_cb (void *data, const char *completion_item,
+irc_completion_server_privates_cb (const void *pointer, void *data,
+                                   const char *completion_item,
                                    struct t_gui_buffer *buffer,
                                    struct t_gui_completion *completion)
 {
@@ -132,9 +146,9 @@ irc_completion_server_privates_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
-    (void) buffer;
 
     if (ptr_server)
     {
@@ -143,8 +157,8 @@ irc_completion_server_privates_cb (void *data, const char *completion_item,
         {
             if (ptr_channel->type == IRC_CHANNEL_TYPE_PRIVATE)
             {
-                weechat_hook_completion_list_add (completion, ptr_channel->name,
-                                                  0, WEECHAT_LIST_POS_SORT);
+                weechat_completion_list_add (completion, ptr_channel->name,
+                                             0, WEECHAT_LIST_POS_SORT);
             }
         }
     }
@@ -157,7 +171,8 @@ irc_completion_server_privates_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_server_nicks_cb (void *data, const char *completion_item,
+irc_completion_server_nicks_cb (const void *pointer, void *data,
+                                const char *completion_item,
                                 struct t_gui_buffer *buffer,
                                 struct t_gui_completion *completion)
 {
@@ -167,6 +182,7 @@ irc_completion_server_nicks_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
@@ -180,15 +196,15 @@ irc_completion_server_nicks_cb (void *data, const char *completion_item,
                 for (ptr_nick = ptr_channel2->nicks; ptr_nick;
                      ptr_nick = ptr_nick->next_nick)
                 {
-                    weechat_hook_completion_list_add (completion, ptr_nick->name,
-                                                      1, WEECHAT_LIST_POS_SORT);
+                    weechat_completion_list_add (completion, ptr_nick->name,
+                                                 1, WEECHAT_LIST_POS_SORT);
                 }
             }
         }
 
         /* add self nick at the end */
-        weechat_hook_completion_list_add (completion, ptr_server->nick,
-                                          1, WEECHAT_LIST_POS_END);
+        weechat_completion_list_add (completion, ptr_server->nick,
+                                     1, WEECHAT_LIST_POS_END);
     }
 
     return WEECHAT_RC_OK;
@@ -199,13 +215,15 @@ irc_completion_server_nicks_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_servers_cb (void *data, const char *completion_item,
+irc_completion_servers_cb (const void *pointer, void *data,
+                           const char *completion_item,
                            struct t_gui_buffer *buffer,
                            struct t_gui_completion *completion)
 {
     struct t_irc_server *ptr_server;
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
     (void) buffer;
@@ -213,8 +231,8 @@ irc_completion_servers_cb (void *data, const char *completion_item,
     for (ptr_server = irc_servers; ptr_server;
          ptr_server = ptr_server->next_server)
     {
-        weechat_hook_completion_list_add (completion, ptr_server->name,
-                                          0, WEECHAT_LIST_POS_SORT);
+        weechat_completion_list_add (completion, ptr_server->name,
+                                     0, WEECHAT_LIST_POS_SORT);
     }
 
     return WEECHAT_RC_OK;
@@ -225,20 +243,22 @@ irc_completion_servers_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_channel_cb (void *data, const char *completion_item,
+irc_completion_channel_cb (const void *pointer, void *data,
+                           const char *completion_item,
                            struct t_gui_buffer *buffer,
                            struct t_gui_completion *completion)
 {
     IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
     if (ptr_channel)
     {
-        weechat_hook_completion_list_add (completion, ptr_channel->name,
-                                          0, WEECHAT_LIST_POS_SORT);
+        weechat_completion_list_add (completion, ptr_channel->name,
+                                     0, WEECHAT_LIST_POS_SORT);
     }
 
     return WEECHAT_RC_OK;
@@ -266,10 +286,10 @@ irc_completion_channel_nicks_add_speakers (struct t_gui_completion *completion,
                 weechat_list_get (channel->nicks_speaking[highlight], i));
             if (nick && irc_nick_search (server, channel, nick))
             {
-                weechat_hook_completion_list_add (completion,
-                                                  nick,
-                                                  1,
-                                                  WEECHAT_LIST_POS_BEGINNING);
+                weechat_completion_list_add (completion,
+                                             nick,
+                                             1,
+                                             WEECHAT_LIST_POS_BEGINNING);
             }
         }
     }
@@ -280,7 +300,8 @@ irc_completion_channel_nicks_add_speakers (struct t_gui_completion *completion,
  */
 
 int
-irc_completion_channel_nicks_cb (void *data, const char *completion_item,
+irc_completion_channel_nicks_cb (const void *pointer, void *data,
+                                 const char *completion_item,
                                  struct t_gui_buffer *buffer,
                                  struct t_gui_completion *completion)
 {
@@ -289,6 +310,7 @@ irc_completion_channel_nicks_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
@@ -300,10 +322,10 @@ irc_completion_channel_nicks_cb (void *data, const char *completion_item,
                 for (ptr_nick = ptr_channel->nicks; ptr_nick;
                      ptr_nick = ptr_nick->next_nick)
                 {
-                    weechat_hook_completion_list_add (completion,
-                                                      ptr_nick->name,
-                                                      1,
-                                                      WEECHAT_LIST_POS_SORT);
+                    weechat_completion_list_add (completion,
+                                                 ptr_nick->name,
+                                                 1,
+                                                 WEECHAT_LIST_POS_SORT);
                 }
                 /* add recent speakers on channel */
                 if (weechat_config_integer (irc_config_look_nick_completion_smart) == IRC_CONFIG_NICK_COMPLETION_SMART_SPEAKERS)
@@ -316,22 +338,22 @@ irc_completion_channel_nicks_cb (void *data, const char *completion_item,
                     irc_completion_channel_nicks_add_speakers (completion, ptr_server, ptr_channel, 1);
                 }
                 /* add self nick at the end */
-                weechat_hook_completion_list_add (completion,
-                                                  ptr_server->nick,
-                                                  1,
-                                                  WEECHAT_LIST_POS_END);
+                weechat_completion_list_add (completion,
+                                             ptr_server->nick,
+                                             1,
+                                             WEECHAT_LIST_POS_END);
                 break;
             case IRC_CHANNEL_TYPE_PRIVATE:
                 /* remote nick */
-                weechat_hook_completion_list_add (completion,
-                                                  ptr_channel->name,
-                                                  1,
-                                                  WEECHAT_LIST_POS_SORT);
+                weechat_completion_list_add (completion,
+                                             ptr_channel->name,
+                                             1,
+                                             WEECHAT_LIST_POS_SORT);
                 /* add self nick at the end */
-                weechat_hook_completion_list_add (completion,
-                                                  ptr_server->nick,
-                                                  1,
-                                                  WEECHAT_LIST_POS_END);
+                weechat_completion_list_add (completion,
+                                             ptr_server->nick,
+                                             1,
+                                             WEECHAT_LIST_POS_END);
                 break;
         }
         ptr_channel->nick_completion_reset = 0;
@@ -345,7 +367,8 @@ irc_completion_channel_nicks_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_channel_nicks_hosts_cb (void *data, const char *completion_item,
+irc_completion_channel_nicks_hosts_cb (const void *pointer, void *data,
+                                       const char *completion_item,
                                        struct t_gui_buffer *buffer,
                                        struct t_gui_completion *completion)
 {
@@ -356,6 +379,7 @@ irc_completion_channel_nicks_hosts_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
@@ -367,10 +391,10 @@ irc_completion_channel_nicks_hosts_cb (void *data, const char *completion_item,
                 for (ptr_nick = ptr_channel->nicks; ptr_nick;
                      ptr_nick = ptr_nick->next_nick)
                 {
-                    weechat_hook_completion_list_add (completion,
-                                                      ptr_nick->name,
-                                                      1,
-                                                      WEECHAT_LIST_POS_SORT);
+                    weechat_completion_list_add (completion,
+                                                 ptr_nick->name,
+                                                 1,
+                                                 WEECHAT_LIST_POS_SORT);
                     if (ptr_nick->host)
                     {
                         length = strlen (ptr_nick->name) + 1 +
@@ -380,7 +404,7 @@ irc_completion_channel_nicks_hosts_cb (void *data, const char *completion_item,
                         {
                             snprintf (buf, length, "%s!%s",
                                       ptr_nick->name, ptr_nick->host);
-                            weechat_hook_completion_list_add (
+                            weechat_completion_list_add (
                                 completion, buf, 0, WEECHAT_LIST_POS_SORT);
                             free (buf);
                         }
@@ -388,7 +412,7 @@ irc_completion_channel_nicks_hosts_cb (void *data, const char *completion_item,
                 }
                 break;
             case IRC_CHANNEL_TYPE_PRIVATE:
-                weechat_hook_completion_list_add (
+                weechat_completion_list_add (
                     completion, ptr_channel->name, 1, WEECHAT_LIST_POS_SORT);
                 break;
         }
@@ -397,12 +421,99 @@ irc_completion_channel_nicks_hosts_cb (void *data, const char *completion_item,
     return WEECHAT_RC_OK;
 }
 
+/*
+ * Adds modelist masks of current channel to completion list.
+ */
+
+int
+irc_completion_modelist_masks_cb (const void *pointer, void *data,
+                                  const char *completion_item,
+                                  struct t_gui_buffer *buffer,
+                                  struct t_gui_completion *completion)
+{
+    char *pos;
+    struct t_irc_modelist *ptr_modelist;
+    struct t_irc_modelist_item *ptr_item;
+
+    IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+
+    pos = strchr (completion_item, ':');
+    if (pos)
+        pos++;
+
+    if (pos && pos[0] && ptr_channel)
+    {
+        ptr_modelist = irc_modelist_search (ptr_channel, pos[0]);
+        if (ptr_modelist)
+        {
+            for (ptr_item = ptr_modelist->items; ptr_item;
+                 ptr_item = ptr_item->next_item)
+            {
+                weechat_completion_list_add (completion,
+                                             ptr_item->mask,
+                                             0, WEECHAT_LIST_POS_END);
+            }
+        }
+    }
+
+    return WEECHAT_RC_OK;
+}
+
+/*
+ * Adds modelist numbers of current channel to completion list.
+ */
+
+int
+irc_completion_modelist_numbers_cb (const void *pointer, void *data,
+                                    const char *completion_item,
+                                    struct t_gui_buffer *buffer,
+                                    struct t_gui_completion *completion)
+{
+    char *pos, str_number[32];
+    struct t_irc_modelist *ptr_modelist;
+    struct t_irc_modelist_item *ptr_item;
+
+    IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+
+    pos = strchr (completion_item, ':');
+    if (pos)
+        pos++;
+
+    if (pos && pos[0] && ptr_channel)
+    {
+        ptr_modelist = irc_modelist_search (ptr_channel, pos[0]);
+        if (ptr_modelist)
+        {
+            for (ptr_item = ptr_modelist->items; ptr_item;
+                 ptr_item = ptr_item->next_item)
+            {
+                snprintf (str_number, sizeof (str_number),
+                          "%d", ptr_item->number + 1);
+                weechat_completion_list_add (completion,
+                                             str_number,
+                                             0, WEECHAT_LIST_POS_END);
+            }
+        }
+    }
+
+    return WEECHAT_RC_OK;
+}
+
 /*
  * Adds topic of current channel to completion list.
  */
 
 int
-irc_completion_channel_topic_cb (void *data, const char *completion_item,
+irc_completion_channel_topic_cb (const void *pointer, void *data,
+                                 const char *completion_item,
                                  struct t_gui_buffer *buffer,
                                  struct t_gui_completion *completion)
 {
@@ -412,6 +523,7 @@ irc_completion_channel_topic_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
@@ -439,9 +551,9 @@ irc_completion_channel_topic_cb (void *data, const char *completion_item,
         else
             topic = strdup (ptr_channel->topic);
 
-        weechat_hook_completion_list_add (completion,
-                                          (topic) ? topic : ptr_channel->topic,
-                                          0, WEECHAT_LIST_POS_SORT);
+        weechat_completion_list_add (completion,
+                                     (topic) ? topic : ptr_channel->topic,
+                                     0, WEECHAT_LIST_POS_SORT);
         if (topic)
             free (topic);
     }
@@ -454,32 +566,71 @@ irc_completion_channel_topic_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_channels_cb (void *data, const char *completion_item,
+irc_completion_channels_cb (const void *pointer, void *data,
+                            const char *completion_item,
                             struct t_gui_buffer *buffer,
                             struct t_gui_completion *completion)
 {
-    struct t_irc_server *ptr_server;
-    struct t_irc_channel *ptr_channel;
+    struct t_irc_server *ptr_server2;
+    struct t_irc_channel *ptr_channel2;
+    struct t_weelist *channels_current_server;
+    int i;
+
+    IRC_BUFFER_GET_SERVER_CHANNEL(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
-    (void) buffer;
 
-    for (ptr_server = irc_servers; ptr_server;
-         ptr_server = ptr_server->next_server)
+    channels_current_server = weechat_list_new ();
+
+    for (ptr_server2 = irc_servers; ptr_server2;
+         ptr_server2 = ptr_server2->next_server)
     {
-        for (ptr_channel = ptr_server->channels; ptr_channel;
-             ptr_channel = ptr_channel->next_channel)
+        for (ptr_channel2 = ptr_server2->channels; ptr_channel2;
+             ptr_channel2 = ptr_channel2->next_channel)
         {
-            if (ptr_channel->type == IRC_CHANNEL_TYPE_CHANNEL)
+            if (ptr_channel2->type == IRC_CHANNEL_TYPE_CHANNEL)
             {
-                weechat_hook_completion_list_add (completion, ptr_channel->name,
-                                                  0, WEECHAT_LIST_POS_SORT);
+                if (ptr_server2 == ptr_server)
+                {
+                    /* will be added later to completions */
+                    weechat_list_add (channels_current_server,
+                                      ptr_channel2->name,
+                                      WEECHAT_LIST_POS_SORT,
+                                      NULL);
+                }
+                else
+                {
+                    weechat_completion_list_add (completion,
+                                                 ptr_channel2->name,
+                                                 0,
+                                                 WEECHAT_LIST_POS_SORT);
+                }
             }
         }
     }
 
+    /* add channels of current server first in list */
+    for (i = weechat_list_size (channels_current_server) - 1; i >= 0; i--)
+    {
+        weechat_completion_list_add (
+            completion,
+            weechat_list_string (
+                weechat_list_get (channels_current_server, i)),
+            0,
+            WEECHAT_LIST_POS_BEGINNING);
+    }
+    weechat_list_free (channels_current_server);
+
+    /* add current channel first in list */
+    if (ptr_channel)
+    {
+        weechat_completion_list_add (completion, ptr_channel->name,
+                                     0, WEECHAT_LIST_POS_BEGINNING);
+    }
+
     return WEECHAT_RC_OK;
 }
 
@@ -488,7 +639,8 @@ irc_completion_channels_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_privates_cb (void *data, const char *completion_item,
+irc_completion_privates_cb (const void *pointer, void *data,
+                            const char *completion_item,
                             struct t_gui_buffer *buffer,
                             struct t_gui_completion *completion)
 {
@@ -496,6 +648,7 @@ irc_completion_privates_cb (void *data, const char *completion_item,
     struct t_irc_channel *ptr_channel;
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
     (void) buffer;
@@ -508,8 +661,8 @@ irc_completion_privates_cb (void *data, const char *completion_item,
         {
             if (ptr_channel->type == IRC_CHANNEL_TYPE_PRIVATE)
             {
-                weechat_hook_completion_list_add (completion, ptr_channel->name,
-                                                  0, WEECHAT_LIST_POS_SORT);
+                weechat_completion_list_add (completion, ptr_channel->name,
+                                             0, WEECHAT_LIST_POS_SORT);
             }
         }
     }
@@ -522,7 +675,8 @@ irc_completion_privates_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_msg_kick_cb (void *data, const char *completion_item,
+irc_completion_msg_kick_cb (const void *pointer, void *data,
+                            const char *completion_item,
                             struct t_gui_buffer *buffer,
                             struct t_gui_completion *completion)
 {
@@ -531,17 +685,18 @@ irc_completion_msg_kick_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
     if (ptr_server)
     {
         msg_kick = IRC_SERVER_OPTION_STRING(ptr_server,
-                                            IRC_SERVER_OPTION_DEFAULT_MSG_KICK);
+                                            IRC_SERVER_OPTION_MSG_KICK);
         if (msg_kick && msg_kick[0])
         {
-            weechat_hook_completion_list_add (completion, msg_kick,
-                                              0, WEECHAT_LIST_POS_SORT);
+            weechat_completion_list_add (completion, msg_kick,
+                                         0, WEECHAT_LIST_POS_SORT);
         }
     }
 
@@ -553,7 +708,8 @@ irc_completion_msg_kick_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_msg_part_cb (void *data, const char *completion_item,
+irc_completion_msg_part_cb (const void *pointer, void *data,
+                            const char *completion_item,
                             struct t_gui_buffer *buffer,
                             struct t_gui_completion *completion)
 {
@@ -562,17 +718,18 @@ irc_completion_msg_part_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
     if (ptr_server)
     {
         msg_part = IRC_SERVER_OPTION_STRING(ptr_server,
-                                            IRC_SERVER_OPTION_DEFAULT_MSG_PART);
+                                            IRC_SERVER_OPTION_MSG_PART);
         if (msg_part && msg_part[0])
         {
-            weechat_hook_completion_list_add (completion, msg_part,
-                                              0, WEECHAT_LIST_POS_SORT);
+            weechat_completion_list_add (completion, msg_part,
+                                         0, WEECHAT_LIST_POS_SORT);
         }
     }
 
@@ -584,7 +741,8 @@ irc_completion_msg_part_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_ignores_numbers_cb (void *data, const char *completion_item,
+irc_completion_ignores_numbers_cb (const void *pointer, void *data,
+                                   const char *completion_item,
                                    struct t_gui_buffer *buffer,
                                    struct t_gui_completion *completion)
 {
@@ -592,6 +750,7 @@ irc_completion_ignores_numbers_cb (void *data, const char *completion_item,
     char str_number[32];
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
     (void) buffer;
@@ -600,8 +759,8 @@ irc_completion_ignores_numbers_cb (void *data, const char *completion_item,
          ptr_ignore = ptr_ignore->next_ignore)
     {
         snprintf (str_number, sizeof (str_number), "%d", ptr_ignore->number);
-        weechat_hook_completion_list_add (completion, str_number,
-                                          0, WEECHAT_LIST_POS_END);
+        weechat_completion_list_add (completion, str_number,
+                                     0, WEECHAT_LIST_POS_END);
     }
 
     return WEECHAT_RC_OK;
@@ -612,7 +771,8 @@ irc_completion_ignores_numbers_cb (void *data, const char *completion_item,
  */
 
 int
-irc_completion_notify_nicks_cb (void *data, const char *completion_item,
+irc_completion_notify_nicks_cb (const void *pointer, void *data,
+                                const char *completion_item,
                                 struct t_gui_buffer *buffer,
                                 struct t_gui_completion *completion)
 {
@@ -621,6 +781,7 @@ irc_completion_notify_nicks_cb (void *data, const char *completion_item,
     IRC_BUFFER_GET_SERVER(buffer);
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
 
@@ -629,8 +790,8 @@ irc_completion_notify_nicks_cb (void *data, const char *completion_item,
         for (ptr_notify = ptr_server->notify_list; ptr_notify;
              ptr_notify = ptr_notify->next_notify)
         {
-            weechat_hook_completion_list_add (completion, ptr_notify->nick,
-                                              0, WEECHAT_LIST_POS_SORT);
+            weechat_completion_list_add (completion, ptr_notify->nick,
+                                         0, WEECHAT_LIST_POS_SORT);
         }
     }
     else
@@ -641,8 +802,8 @@ irc_completion_notify_nicks_cb (void *data, const char *completion_item,
             for (ptr_notify = ptr_server->notify_list; ptr_notify;
                  ptr_notify = ptr_notify->next_notify)
             {
-                weechat_hook_completion_list_add (completion, ptr_notify->nick,
-                                                  0, WEECHAT_LIST_POS_SORT);
+                weechat_completion_list_add (completion, ptr_notify->nick,
+                                             0, WEECHAT_LIST_POS_SORT);
             }
         }
     }
@@ -650,6 +811,62 @@ irc_completion_notify_nicks_cb (void *data, const char *completion_item,
     return WEECHAT_RC_OK;
 }
 
+/*
+ * Adds filters for raw buffer to completion list.
+ */
+
+int
+irc_completion_raw_filters_cb (const void *pointer, void *data,
+                               const char *completion_item,
+                               struct t_gui_buffer *buffer,
+                               struct t_gui_completion *completion)
+{
+    char str_filter[1024];
+
+    IRC_BUFFER_GET_SERVER(buffer);
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) completion_item;
+
+    /* all messages */
+    weechat_completion_list_add (completion, "*",
+                                 0, WEECHAT_LIST_POS_SORT);
+
+    /* condition */
+    weechat_completion_list_add (completion,
+                                 "c:${recv} && ${command}==PRIVMSG",
+                                 0, WEECHAT_LIST_POS_SORT);
+
+    /* message flag */
+    weechat_completion_list_add (completion, "f:modified",
+                                 0, WEECHAT_LIST_POS_SORT);
+    weechat_completion_list_add (completion, "f:recv",
+                                 0, WEECHAT_LIST_POS_SORT);
+    weechat_completion_list_add (completion, "f:redirected",
+                                 0, WEECHAT_LIST_POS_SORT);
+    weechat_completion_list_add (completion, "f:sent",
+                                 0, WEECHAT_LIST_POS_SORT);
+
+    /* IRC command */
+    weechat_completion_list_add (completion, "m:notice",
+                                 0, WEECHAT_LIST_POS_SORT);
+    weechat_completion_list_add (completion, "m:privmsg",
+                                 0, WEECHAT_LIST_POS_SORT);
+
+    /* server */
+    for (ptr_server = irc_servers; ptr_server;
+         ptr_server = ptr_server->next_server)
+    {
+        snprintf (str_filter, sizeof (str_filter), "s:%s", ptr_server->name);
+        weechat_completion_list_add (completion, str_filter,
+                                     0, WEECHAT_LIST_POS_SORT);
+    }
+
+    return WEECHAT_RC_OK;
+}
+
 /*
  * Hooks completions.
  */
@@ -659,50 +876,61 @@ irc_completion_init ()
 {
     weechat_hook_completion ("irc_server",
                              N_("current IRC server"),
-                             &irc_completion_server_cb, NULL);
+                             &irc_completion_server_cb, NULL, NULL);
     weechat_hook_completion ("irc_server_nick",
                              N_("nick on current IRC server"),
-                             &irc_completion_server_nick_cb, NULL);
+                             &irc_completion_server_nick_cb, NULL, NULL);
     weechat_hook_completion ("irc_server_channels",
                              N_("channels on current IRC server"),
-                             &irc_completion_server_channels_cb, NULL);
+                             &irc_completion_server_channels_cb, NULL, NULL);
     weechat_hook_completion ("irc_server_privates",
                              N_("privates on current IRC server"),
-                             &irc_completion_server_privates_cb, NULL);
+                             &irc_completion_server_privates_cb, NULL, NULL);
     weechat_hook_completion ("irc_server_nicks",
                              N_("nicks on all channels of current IRC server"),
-                             &irc_completion_server_nicks_cb, NULL);
+                             &irc_completion_server_nicks_cb, NULL, NULL);
     weechat_hook_completion ("irc_servers",
                              N_("IRC servers (internal names)"),
-                             &irc_completion_servers_cb, NULL);
+                             &irc_completion_servers_cb, NULL, NULL);
     weechat_hook_completion ("irc_channel",
                              N_("current IRC channel"),
-                             &irc_completion_channel_cb, NULL);
+                             &irc_completion_channel_cb, NULL, NULL);
     weechat_hook_completion ("nick",
                              N_("nicks of current IRC channel"),
-                             &irc_completion_channel_nicks_cb, NULL);
+                             &irc_completion_channel_nicks_cb, NULL, NULL);
     weechat_hook_completion ("irc_channel_nicks_hosts",
                              N_("nicks and hostnames of current IRC channel"),
-                             &irc_completion_channel_nicks_hosts_cb, NULL);
+                             &irc_completion_channel_nicks_hosts_cb, NULL, NULL);
+    weechat_hook_completion ("irc_modelist_masks",
+                             N_("modelist masks of current IRC channel; "
+                                "required argument: modelist mode"),
+                             &irc_completion_modelist_masks_cb, NULL, NULL);
+    weechat_hook_completion ("irc_modelist_numbers",
+                             N_("modelist numbers of current IRC channel; "
+                                "required argument: modelist mode"),
+                             &irc_completion_modelist_numbers_cb, NULL, NULL);
     weechat_hook_completion ("irc_channel_topic",
                              N_("topic of current IRC channel"),
-                             &irc_completion_channel_topic_cb, NULL);
+                             &irc_completion_channel_topic_cb, NULL, NULL);
     weechat_hook_completion ("irc_channels",
                              N_("channels on all IRC servers"),
-                             &irc_completion_channels_cb, NULL);
+                             &irc_completion_channels_cb, NULL, NULL);
     weechat_hook_completion ("irc_privates",
                              N_("privates on all IRC servers"),
-                             &irc_completion_privates_cb, NULL);
+                             &irc_completion_privates_cb, NULL, NULL);
     weechat_hook_completion ("irc_msg_kick",
                              N_("default kick message"),
-                             &irc_completion_msg_kick_cb, NULL);
+                             &irc_completion_msg_kick_cb, NULL, NULL);
     weechat_hook_completion ("irc_msg_part",
                              N_("default part message for IRC channel"),
-                             &irc_completion_msg_part_cb, NULL);
+                             &irc_completion_msg_part_cb, NULL, NULL);
     weechat_hook_completion ("irc_ignores_numbers",
                              N_("numbers for defined ignores"),
-                             &irc_completion_ignores_numbers_cb, NULL);
+                             &irc_completion_ignores_numbers_cb, NULL, NULL);
     weechat_hook_completion ("irc_notify_nicks",
                              N_("nicks in notify list"),
-                             &irc_completion_notify_nicks_cb, NULL);
+                             &irc_completion_notify_nicks_cb, NULL, NULL);
+    weechat_hook_completion ("irc_raw_filters",
+                             N_("filters for irc raw buffer"),
+                             &irc_completion_raw_filters_cb, NULL, NULL);
 }
